/* This software is licensed by the MIT License, see LICENSE file */
/*                                Copyright © 2022 Gregory Lirent */

#include "../modules/libcdsb/modules/libunic/include.h"
#include "buffer.h"

#define unescape8  libcjsonp_builtin_unescape8
#define unescape16 libcjsonp_builtin_unescape32
#define unescape32 libcjsonp_builtin_unescape16

static bool libcjsonp_builtin_unescape8(unsigned char* x, const char* s) {
    unsigned char c[2];

    if (*s > 0x39) {
        c[0] = (*s|0x20);

        if (c[0] < 0x61 || c[0] > 0x66)
            return false;

        c[0] -= 0x57;
    } else if (*s >= 0x30) {
        c[0] = *s&0x0f;
    } else return false;

    if (*++s > 0x39) {
        c[1] = (*s|0x20);

        if (c[1] < 0x61 || c[1] > 0x66)
            return false;

        c[1] -= 0x57;
    } else if (*s >= 0x30) {
        c[1] |= *s&0x0f;
    } else return false;

    *x = (c[0]<<4) | c[1];

    return true;
}

static void libcjsonp_builtin_unescape32(char** x, char** s) {

    vtype_uint32 uc;
    size_t n;
    bool   f;
    char*  p;

    n  =  4;
    uc =  0;
    f  =  0;
    p  = *s;

    while (n--) {
        unsigned char c;

        f |= !libcjsonp_builtin_unescape8(&c, p);
        uc = c | uc >> 8;

        p += 2;
    }

    if (!f) {
        *s = p;
        if (!f && (p = tochar_unicode(*x, uc)))
            *x = p;
    } else --*s;
}

static void libcjsonp_builtin_unescape16(char** x, char** s) {

    vtype_uint32 uc;
    char*  p;
    bool   f;
    unsigned char c[2];

    f =  0;
    p = *s;

    f |= !libcjsonp_builtin_unescape8(c, p);
    f |= !libcjsonp_builtin_unescape8(c + 1, p += 2);

    p += 2;

    if (f) goto bad_;

    if (*c < 0xd8 || *c >= 0xdc) {
        uc = *c << 8 | c[1];
        goto end_;
    }

    if ((p)[0] != '\\' || (p)[1] != 'u')
        return;

    p += 2;

    uc  = (c[0]&0x03) << 18;
    uc |= c[1] << 10;

    f |= !libcjsonp_builtin_unescape8(c, p);
    f |= !libcjsonp_builtin_unescape8(c + 1, p += 2);

    p += 2;

    if (!f) {
        if (*c < 0xdc || *c > 0xdf)
            goto bad_;

        uc |= (c[0]&0x03)<<8;
        uc |= c[1];
        uc += 0x010000;

        end_: *s = p;
        if ((p = tochar_unicode(*x, uc)))
            *x = p;
    } else bad_: --*s;
}

/*#####################################################################################################################*/

void libcjsonp_string_unescape(vtype_string* x) {

    char *p, *c;

    if (!x->buffer || !*x->buffer)
        return;

    p = c = x->buffer;

    while (*p) {

        if (*p == '\\') {
            switch (*(++p)) {
                default :   *(c++) =   *p; break;
                case 'n':   *(c++) = '\n'; break;
                case 'r':   *(c++) = '\r'; break;
                case 'e':   *(c++) = '\e'; break;
                case 'v':   *(c++) = '\v'; break;
                case 't':   *(c++) = '\t'; break;
                case 'a':   *(c++) = '\a'; break;

                case 'x':   if (unescape8((void*)c, p + 1)) {
                                ++c;
                                p += 3;
                            } goto end_loop_;
                case 'U': ++p; unescape32(&c, &p); goto end_loop_;
                case 'u': ++p; unescape16(&c, &p); goto end_loop_;
            }

        } else if (p != c) {
            *(c++) = *p;
        } else ++c;

        ++p;

        end_loop_:{}
    }

    *c = 0;
}


void libcjsonp_string_escape(vtype_string* s) {
    buffer_t x;

    if (!s->buffer || !*s->buffer)
        return;

    memset(&x, 0, sizeof(x));

    buffer_write_and_wrap(&x, s->buffer);

    x.n       = (x.ptr - x.mem) - 1;
    s->buffer = libcdsb_realloc(s->buffer, x.n);
    x.ptr[-1] = 0;

    memcpy(s->buffer, x.mem + 1, x.n);
    libcdsb_free(x.mem);
}
